// qmSubMatrixRef.c++

#include "qmSubMatrixRef.h"
#include "qmMatrixRef.h"
#include "qmMatrix.h"
#include <memory.h>

qmSubMatrixRef::qmSubMatrixRef( const qmMatrixRef& m, 
				int h, int w, int r, int c )
  : elementCount( ((h<=0)?m.getHeight():h) * ((w<=0)?m.getWidth():w) ),
    width( ((w<=0)?m.getWidth():w) ),
    spacing( m.getWidth() - w ),
    array( m._getArray(r,c) )
{}

qmSubMatrixRef&
qmSubMatrixRef::setData( real *p )
{
    register real *to = array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      *to++ = *p++;
	   to += spacing;
       }
    return *this;
}

real *
qmSubMatrixRef::getData( real *p ) const
{
    register real *fm = array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      *p++ = *fm++;
	   fm += spacing;
       }
    return p;
}

qmSubMatrixRef&
qmSubMatrixRef::setAdd( const qmSubMatrixRef& a, const qmSubMatrixRef& b )
{
    real *to = array;
    real *fma = a.array;
    real *fmb = b.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      (*to++) = (*fma++) + (*fmb++);
	   to += spacing;
	   fma += a.spacing;
	   fmb += b.spacing;
       }
    return *this;
}

qmSubMatrixRef&
qmSubMatrixRef::setSubtract( const qmSubMatrixRef& a, const qmSubMatrixRef& b )
{
    real *to = array;
    real *fma = a.array;
    real *fmb = b.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      (*to++) = (*fma++) - (*fmb++);
	   to += spacing;
	   fma += a.spacing;
	   fmb += b.spacing;
       }
    return *this;
}

qmSubMatrixRef&
qmSubMatrixRef::setMultiply( const qmSubMatrixRef& a, const qmSubMatrixRef& b )
{
    register real *to = array;
    register real *fma;
    register real *fmb;
    int height = elementCount / width;
    int bline = b.width + b.spacing;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      {
		  real sum = 0.;
		  fma = a.getArray( i, 0 );
		  fmb = b.getArray( 0, j );
		  for( int k = 0; k < a.width; k++ )
		     sum += (*fma++) * (*fmb), fmb += bline;
		  *to++ = sum;
	      }
	   to += spacing;
       }
    return *this;
}

qmSubMatrixRef&
qmSubMatrixRef::setMultiplyT( const qmSubMatrixRef& a, const qmSubMatrixRef& b )
{
    register real *to = array;
    register real *fma;
    register real *fmb;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      {
		  real sum = 0.;
		  fma = a.getArray( i, 0 );
		  fmb = b.getArray( j, 0 );
		  for( int k = 0; k < a.width; k++ )
		     sum += (*fma++) * (*fmb++);
		  *to++ = sum;
	      }
	   to += spacing;
       }
    return *this;
}

qmSubMatrixRef&
qmSubMatrixRef::setAddEq( const qmSubMatrixRef& a )
{
    real *to = array;
    real *fm = a.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      (*to++) += (*fm++);
	   to += spacing;
	   fm += a.spacing;
       }
    return *this;
}

qmSubMatrixRef&
qmSubMatrixRef::setInvert( const qmSubMatrixRef& from )
{
    assert( getWidth() == from.getWidth() && getHeight() == from.getHeight() &&
	    getWidth() == getHeight() );
    
    qmMatrix copy( elementCount/width, width );
    from.getData( copy.getArray() );
    copy.setInvertEq();
    setData( copy.getArray() );
    return *this;
}

qmSubMatrixRef&
qmSubMatrixRef::setInvertEq()
{
    return setInvert( *this );
}

qmSubMatrixRef&
qmSubMatrixRef::setScaleEq( real val )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       (*to++) *= val, to += spacing;
    return *this;
}

qmSubMatrixRef&
qmSubMatrixRef::setConst( real val )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       (*to++) = val, to += spacing;
    return *this;
}

qmSubMatrixRef&
qmSubMatrixRef::setDiagonal( real val )
{
    int height = elementCount / width;
    int count = ( width < height ) ? width : height;
    setZero();
    while( count-- )
       (*this)( count, count ) = val;
    return *this;
}

ostream&
qmSubMatrixRef::print( ostream& os, char *lineend )
{
    register int count = elementCount;
    register real *fm = array;
    os << "e= " << elementCount << ", w= " << width 
       << ", a= " << array << endl;
    for( ; count; count -= width )
       {
	   for( int i = 0; i < width; i++ )
	      os << " " << *fm++, fm += spacing;
	   os << ((lineend)?lineend:"\0");
       }
    return os;
}

istream&
qmSubMatrixRef::scan( istream& is )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       is >> *to++, to += spacing;
    return is;
}

ostream&
operator<<( ostream& os, qmSubMatrixRef& sm )
{
    return sm.print( os );
}


qmSubMatrixRef&
qmSubMatrixRef::copy( const qmSubMatrixRef& sm )
{
    assert( elementCount == sm.elementCount );
    width = sm.width;
    memcpy( array, sm.array, elementCount * sizeof(real) );
    return *this;
}


ostream&
qmSubMatrixRef::write( ostream& os )
{
    return( os.write( (char *)array, elementCount * sizeof(real) ) );
}


istream&
qmSubMatrixRef::read( istream& is )
{
    return( is.read( (char *)array, elementCount * sizeof(real) ) );
}

