// qmVectorRef.h++

#ifndef _qmVectorRef_H_
#define _qmVectorRef_H_

#include <stddef.h>
#include <iostream.h>
#include <assert.h>
#include <math.h>
#include "qmReal.h"
#include "qmMatrixRef.h"

class qmVectorRef
{
    friend class qmMatrixRef;
    friend class qmVector;
    
protected:
    
    real *array;
    int elementCount;
    
public:
    
    virtual ~qmVectorRef() {}
    qmVectorRef( )
      : array(0), elementCount(0) {}
    qmVectorRef( int n, real* a )
      : array(a), elementCount(n) {}
    qmVectorRef( const qmVectorRef& vr )
      : array( 0 ), elementCount( 0 ) { set( vr ); }

    qmVectorRef& set( int n, real* a );			// shadow
    qmVectorRef& set( const qmVectorRef& );		// shadow
    qmVectorRef& copy( const qmVectorRef& );		// copy
    qmVectorRef& operator=( const qmVectorRef& );	// shadow

    int getLength() const { return elementCount; }
    int getElementCount() const { return elementCount; }

    real* getArray( int i = 0 ) { return( array + i ); }
    operator real*() { return array; }
    real& operator()( int i ) { return array[i]; }
    real getValue( int i ) const { return array[i]; }
    inline qmVectorRef subVector( int i, int n ) const;

    void setData( const real *p );
    void getData( real *p ) const;
    qmVectorRef& operator<<( real *p ) { setData(p); return *this; }
    void operator>>( real *p ) { getData(p); }

    qmVectorRef& setAdd( const qmVectorRef&, const qmVectorRef& );
    qmVectorRef& setSubtract(const qmVectorRef&, const qmVectorRef& );
    qmVectorRef& setNegate( const qmVectorRef& );
    qmVectorRef& setShift( real, const qmVectorRef& );
    qmVectorRef& setScale( real, const qmVectorRef& );
    qmVectorRef& setScaleAddEq( real, const qmVectorRef& );
    qmVectorRef& setAddEq( const qmVectorRef& op );
    qmVectorRef& setSubtractEq( const qmVectorRef& op );
    qmVectorRef& setNegateEq();
    qmVectorRef& setShiftEq( real val );
    qmVectorRef& setScaleEq( real val );
    qmVectorRef& setConst( real val );
    qmVectorRef& setZero() { return setConst( 0.); }
    qmVectorRef& operator+=(const qmVectorRef& op) {return setAddEq(op); }
    qmVectorRef& operator-=(const qmVectorRef& op) {return setSubtractEq(op);}
    qmVectorRef& operator+=(const real val) { return setShiftEq(val); }
    qmVectorRef& operator-=(const real val) { return setShiftEq(-val); }
    qmVectorRef& operator*=(const real val) { return setScaleEq(val); }
    qmVectorRef& operator/=(const real val) { return setScaleEq(real(1)/val); }
    qmVectorRef& operator=( const real val ) { return setConst(val); }
    qmVectorRef& setMultiply( const qmMatrixRef&, const qmVectorRef& );
    qmVectorRef& setTMultiply( const qmMatrixRef&, const qmVectorRef& );
    qmVectorRef& setVectorScale( const qmVectorRef&, const qmVectorRef& );
    real getInnerProduct( const qmVectorRef& ) const;
    int operator==( const qmVectorRef& ) const;
    qmVectorRef& setAbsolute();
    ostream& print( ostream& ) const;
    istream& scan( istream& );
    ostream& write( ostream& ) const;
    istream& read( istream& );
    qmVectorRef& referTo( void* p ) { array = (real*)p; return *this; }
    int getByteCount() { return( elementCount * sizeof(real) ); }
    inline real getMax( int* which = 0 ) const;
    inline real getMin( int* which = 0 ) const;
    inline int setNormalize();
    qmVectorRef& setCross( const qmVectorRef&, const qmVectorRef& );

    inline real getSum() const;

private:
    inline real* _getArray( int i=0 ) const { return( array + i ); }
    inline real& _el( int i ) const { return array[i]; }
};

inline
qmVectorRef
qmVectorRef::subVector( int i, int n ) const
{
    return qmVectorRef( n, array+i );
}

inline
real
qmVectorRef::getMax( int* which ) const
{
#ifdef QmWarnings
    // should make sure array is valid !!
#endif
    real maxv = array[0];
    int maxi = 0;
    for( int i = 1; i < elementCount; i++ )
       if( array[i] > maxv )
	  maxi = i, maxv = array[i];
    if( which )
       *which = maxi;
    return maxv;
}

inline
real
qmVectorRef::getMin( int* which ) const
{
#ifdef QmWarnings
    // should make sure array is valid !!
#endif
    real minv = array[0];
    int mini = 0;
    for( int i = 1; i < elementCount; i++ )
       if( array[i] < minv )
	  mini = i, minv = array[i];
    if( which )
       *which = mini;
    return minv;
}

inline
int
qmVectorRef::setNormalize()
{
    real mag = getInnerProduct( *this );
    mag = (real)sqrt( mag );
    if( mag < .00001 )
       {
	   int which;
	   getMax( &which );
	   setZero();
	   array[which] = 1.;
	   return 1;
       }
    setScaleEq((real)( 1./mag ));
    return 0;
}

inline
real
qmVectorRef::getSum() const
{
    real sum = 0;
    for( int i = 0; i < elementCount; i++ )
       sum += array[i];
    return sum;
}


extern ostream& operator<<( ostream&, const qmVectorRef& );
extern istream& operator>>( istream&, qmVectorRef& );

#endif /* _qmVectorRef_H_ */
