// qmdLUDecomposition.c++

#include "qmdLUDecomposition.h"

static inline reald absVal( reald x ) { return( ( x < 0. ) ? -x : x ); }
const reald SMALLFLOAT = 1.0E-14;

int
qmdLUDecomposition::findMaxPivot( qmdMatrixRef& A, int opRow, int *rowOrder )
{
    // We are trying to find the "opRow"th row, having already processed
    // rows 0 through opRow-1
    // This function finds the row with the largest value in column "opRow"
    // and puts that row next in the rowOrder table
    // This is an inner loop of decompose and does not realdly make
    // sense on its own

    int height = A.getHeight();
    int maxRow = opRow;
    reald maxValue = 0.;
    for ( int i = opRow; i < height; i++) 
       {
	   reald cmpval = absVal( A( rowOrder[i], opRow ) );
	   if ( cmpval > maxValue ) 
	      {
		  maxRow = i;
		  maxValue = cmpval;
	      }
       }
    if ( maxRow != opRow ) 
       {
	   int tmp_i = rowOrder[ opRow ];
	   rowOrder[ opRow ] = rowOrder[ maxRow ];
	   rowOrder[ maxRow ] = tmp_i;
       }
    if ( maxValue < SMALLFLOAT)
       return 1;
    return 0;
}

int
qmdLUDecomposition::LUDecompose( qmdMatrixRef& A, int *rowOrder )
{
    int height = A.getHeight();
    
    /* for each row, zero all elements below that diagonal point */

    for (int i = 0; i < height - 1; i++) 
       {
	   if( findMaxPivot( A, i, rowOrder ) ) 
	      {
#ifdef qmdWarnings
		  cerr << "Singular qmdLUDecomposition?? (aborted)" << endl;
#endif
		  return 1; // Singular??
	      }
	   reald *pivotSpot = A.getArray( rowOrder[i], i );
	   reald inversePivot = 1. / *pivotSpot;
	   for ( int row = i+1; row < height; row++ ) 
	      {
		  reald* multiplierSpot = A.getArray( rowOrder[row], i );
		  *multiplierSpot *= inversePivot;
		  reald multiplier = *multiplierSpot++; // watch it! (++) 
		  reald* pivotRow = pivotSpot + 1;
		  for( int col = i + 1; col < height; col++ )
		     *multiplierSpot++ -= *pivotRow++ * multiplier;
	      }
       }
    return 0;
}

void
qmdLUDecomposition::forwardSubstitute( qmdMatrixRef& LU,
				      int *rowOrder,
				      reald *b,
				      reald *y )

{
    // computes y for Ly = b
    int width = LU.getWidth();
    for ( int i = 0; i < width; i++ ) 
       {
	   int row = rowOrder[i];
	   reald* rowptr = LU.getArray( row, 0 );
	   y[i] = b[row];
	   for ( int col = 0; col < i; col++ )
	      y[i] -= *rowptr++ * y[col];
       }
}

void
qmdLUDecomposition::backwardSubstitute( qmdMatrixRef& LU,
				       int* rowOrder,
				       reald *y,
				       reald *x )
{
    // computes x for Ux = y
    int height = LU.getWidth();
    for ( int i = height-1; i >= 0; i-- ) // each row
       {
	   int row = rowOrder[i];
	   reald* rowptr = LU.getArray( row, i );
	   reald pivot = 1. / *rowptr++;
	   reald sum = 0.;
	   for ( int col = i + 1; col < height; col++)
	      sum += *rowptr++ * x[col];
	   x[i] = (y[i] - sum) * pivot;
       }
}

// qmdLUDecomposition&
int
qmdLUDecomposition::set( const qmdMatrixRef& m )
{
    LUmatrix.set( m );
    if( rowOrder )
       delete [] rowOrder;
    rowOrder = new int[ m.getWidth() ];
    for( int i = 0; i < m.getWidth(); i++ )
       rowOrder[i] = i;
    return LUDecompose( LUmatrix, rowOrder );
    // return *this;
}

int
qmdLUDecomposition::LUSolve( qmdMatrixRef& LU, 
			    int* rorder,
			    qmdVectorRef& y,
			    qmdVectorRef& x )
{
    qmdVector c( y.getLength() );
    forwardSubstitute( LU, rorder, y, c );
    backwardSubstitute( LU, rorder, c, x );
    return 0;
}

int
qmdLUDecomposition::LUInvert( qmdMatrixRef& LU, int* rorder, qmdMatrixRef& Ai )
{
    // use the A=LU decomp to compute Ai (inverse of A)
    int width = LU.getWidth();
    qmdVector y( width );
    qmdVector c( width );
    qmdVector x( width );
    y.setZero();
    for( int i = 0; i < width; i++ )
       {
	   y(i) = reald(1.);
	   forwardSubstitute( LU, rorder, y, c );
	   backwardSubstitute( LU, rorder, c, x );
	   y(i) = reald(0.);
	   for( int j = 0; j < width; j++ )
	      Ai( j, i ) = x( j );
       }
    return 0;
}

int
qmdLUDecomposition::solve( qmdVectorRef& y, qmdVectorRef& x )
{
    return LUSolve( LUmatrix, rowOrder, y, x );
}

int
qmdLUDecomposition::invert( qmdMatrixRef& Ai )
{
    return LUInvert( LUmatrix, rowOrder, Ai );
}

ostream&
qmdLUDecomposition::print( ostream& os )
{
    os << "LU= " << endl << LUmatrix << endl;
    os << "row order=";
    for( int i = 0; i < LUmatrix.getWidth(); i++ )
       os << " " << rowOrder[i];
    os << endl;
    return os;
}

ostream&
operator<<( ostream& os, qmdLUDecomposition& l )
{
    return l.print( os );
}
