// qmdMatrixRef.c++

#include <math.h>
#include "qmdMatrixRef.h"
#include "qmdMatrix.h"
#include "qmdCompileOptions.h"
#include "qmdMinMax.h"
#include "qmdLUDecomposition.h"
#include "qmdEigenStructure.h"
#include <memory.h>

qmdMatrixRef&
qmdMatrixRef::set( int h, int w, reald* a )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( h > 0 && w > 0 && a );
#endif
    elementCount = h*w;
    width = w;
    array = a;
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::set( const qmdMatrixRef& mr )
{
    memcpy( this, &mr, sizeof( qmdMatrixRef ) );
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::copy( const qmdMatrixRef& mr )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount && 
	      elementCount == mr.elementCount &&
	      width == mr.width );
#endif
    memcpy( array, mr.array, elementCount * sizeof(reald) );
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::operator=( const qmdMatrixRef& mr )
{
#ifdef qmdWarnings
    cerr << "qmdMatrixRef: WARNING: operator=() does shadow, not copy" << endl;
    cerr << "                      use set or copy instead" << endl;
#endif
    return set( mr );
}

void
qmdMatrixRef::setData( reald *p )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( p != (reald *) NULL);
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       (*to++) = *p++;
}

void
qmdMatrixRef::getData( reald *p ) const
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( p != (reald *) NULL);
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *fm = array;
    register reald *to = p;
    while( count-- )
       (*to++) = *fm++;
}

qmdMatrixRef&
qmdMatrixRef::setAdd( const qmdMatrixRef& a, const qmdMatrixRef& b )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    elementCount == a.elementCount &&
	    width == a.width &&
	    elementCount == b.elementCount &&
	    width == b.width );
#endif
    reald *to = array;
    reald *fma = a.array;
    reald *fmb = b.array;
    int count = elementCount;
    while( count-- )
       (*to++) = (*fma++) +  (*fmb++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setSubtract( const qmdMatrixRef& a, const qmdMatrixRef& b )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    elementCount == a.elementCount &&
	    width == a.width &&
	    elementCount == b.elementCount &&
	    width == b.width );
#endif
    reald *to = array;
    reald *fma = a.array;
    reald *fmb = b.array;
    int count = elementCount;
    while( count-- )
       (*to++) = (*fma++) -  (*fmb++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setMultiply( const qmdMatrixRef& a, const qmdMatrixRef& b )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    width == b.width &&
	    a.width * b.width == b.elementCount &&
	    elementCount * a.width == a.elementCount * width );
#endif
    register reald *to = array;
    register reald *fma = a.array;
    register reald *fmb = b.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       for( int j = 0; j < width; j++ )
	  {
	      reald sum = 0.;
	      fma = a._getArray( i, 0 );
	      fmb = b._getArray( 0, j );
	      for( int k = 0; k < a.width; k++ )
		 sum += (*fma++) * (*fmb), fmb += b.width;
	      *to++ = sum;
	  }
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setMultiplyT( const qmdMatrixRef& a, const qmdMatrixRef& b )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    width > 0 && a.width > 0 && b.width > 0 &&
	    getHeight() == a.getHeight() &&
	    width == b.getHeight() &&
	    a.width == b.width );
#endif
    register reald *to = array;
    register reald *fma = a.array;
    register reald *fmb = b.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       for( int j = 0; j < width; j++ )
	  {
	      reald sum = 0.;
	      fma = a._getArray( i, 0 );
	      fmb = b._getArray( j, 0 );
	      for( int k = 0; k < a.width; k++ )
		 sum += (*fma++) * (*fmb++);
	      *to++ = sum;
	  }
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setTMultiply( const qmdMatrixRef& a, const qmdMatrixRef& b )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    width > 0 && a.width > 0 && b.width > 0 &&
	    getHeight() == a.width &&
	    width == b.width &&
	    a.getHeight() == b.getHeight() );
#endif
    register reald *to = array;
    register reald *fma = a.array;
    register reald *fmb = b.array;
    int height = elementCount / width;
    int inner = a.getHeight();
    for( int i = 0; i < height; i++ )
       for( int j = 0; j < width; j++ )
	  {
	      reald sum = 0.;
	      fma = a._getArray( 0, i );
	      fmb = b._getArray( 0, j );
	      for( int k = 0; k < inner; k++ )
		 sum += (*fma) * (*fmb), fma+=a.width, fmb+=b.width;
	      *to++ = sum;
	  }
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setDivide( const qmdMatrixRef& a, const qmdMatrixRef& b )
{
#ifdef qmdDoBoundaryChecking
#endif
    qmdMatrix c;
    c.setInvert( b );
    return setMultiply( a, c );
}

qmdMatrixRef&
qmdMatrixRef::setNegate( const qmdMatrixRef& a )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width == a.width );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = a.array;
    while( count-- )
       (*to++) = -(*fm++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setTranspose( const qmdMatrixRef& a )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width * a.width == elementCount );
#endif
    register reald *to = array;
    register reald *fm = a.array;
    int row = 0;
    int col = 0;
    for( int el = 0; el < elementCount; el++ )
       {
	   *to++ = *fm, fm += a.width;
	   if( ++col == width )
	      col = 0, fm = a.array + (++row);
       }
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setShift( reald val, const qmdMatrixRef& a )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    elementCount == a.elementCount &&
	    width == a.width );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = a.array;
    while( count-- )
       (*to++) = val + (*fm++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setScale( reald val, const qmdMatrixRef& a )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    elementCount == a.elementCount &&
	    width == a.width );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = a.array;
    while( count-- )
       (*to++) = val * (*fm++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setScaleAddEq( reald val, const qmdMatrixRef& a )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width == a.width );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = a.array;
    while( count-- )
       (*to++) += val * (*fm++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setAddEq( const qmdMatrixRef& a )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width == a.width );
#endif
    reald *to = array;
    reald *fm = a.array;
    int count = elementCount;
    while( count-- )
       (*to++) += (*fm++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setSubtractEq( const qmdMatrixRef& a )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width == a.width );
#endif
    reald *to = array;
    reald *fm = a.array;
    int count = elementCount;
    while( count-- )
       (*to++) -= (*fm++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setNegateEq()
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    reald *to = array;
    reald *fm = array;
    int count = elementCount;
    while( count-- )
       (*to++) = -(*fm++);
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setInvert( const qmdMatrixRef& from )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    elementCount == from.elementCount &&
	    width == from.width &&
	    width*width == elementCount );
#else
    qmdAssert( getWidth() == from.getWidth() && 
	      getHeight() == from.getHeight() &&
	      getWidth() == getHeight() );
#endif
    // qmdLUD lud( from );
    // lud.invert( *this );
    qmdLUD lud;
    if( lud.set( from ) )
       cerr << "qmdMatrixRef::setInvert: singular inversion (aborted)" << endl;
    else
       lud.invert( *this );
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setInvertEq()
{
#ifdef qmdDoBoundaryChecking
    // check done in setInvert()
#endif
    return setInvert( *this );
}

int
qmdMatrixRef::setInvertCheck( const qmdMatrixRef& from )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	    elementCount == from.elementCount &&
	    width == from.width &&
	    width*width == elementCount );
#else
    qmdAssert( getWidth() == from.getWidth() && 
	      getHeight() == from.getHeight() &&
	      getWidth() == getHeight() );
#endif
    qmdLUD lud;
    if( lud.set( from ) )
       return 1;
    lud.invert( *this );
    return 0;
}

int
qmdMatrixRef::setInvertEqCheck()
{
#ifdef qmdDoBoundaryChecking
    // check done in setInvert()
#endif
    return setInvertCheck( *this );
}

qmdMatrixRef&
qmdMatrixRef::setScaleEq( reald val )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       (*rp++) *= val;
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setConst( reald val )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       (*rp++) = val;
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setDiagonal( reald val )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    int height = elementCount / width;
    int count = ( width < height ) ? width : height;
    setZero();
    while( count-- )
       (*this)( count, count ) = val;
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setDiagonal( const qmdVectorRef& vr )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 && width == vr.elementCount );
#endif
    setZero();
    for( int i = 0; i < width; i++ )
       _el( i, i ) = vr._el( i );
    return *this;
}

void
qmdMatrixRef::getDiagonal( const qmdVectorRef& vr )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 && width == vr.elementCount );
#endif
    for( int i = 0; i < width; i++ )
       vr._el( i ) = _el( i, i );
}

qmdMatrixRef&
qmdMatrixRef::setOuterProduct( const qmdVectorRef& v1, const qmdVectorRef& v2 )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      width == v2.getLength() &&
	      width * v1.getLength() == elementCount );
#endif
    int count = elementCount;
    reald* to = array;
    reald* fmrow = v1._getArray();
    reald* fmcol = v2._getArray();
    int col = 0;
    while( count-- )
       {
	   *to++ = *fmrow * (*fmcol++);
	   if( ++col == width )
	      col = 0, fmrow++, fmcol = v2._getArray();
       }
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setAddScaledOuterProductEq( const qmdVectorRef& v1,
					 const qmdVectorRef& v2,
					 reald w )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      width == v2.getLength() &&
	      width * v1.getLength() == elementCount );
#endif
    int count = elementCount;
    reald* to = array;
    reald* fmrow = v1._getArray();
    reald* fmcol = v2._getArray();
    int col = 0;
    while( count-- )
       {
	   *to++ += w * (*fmrow) * (*fmcol++);
	   if( ++col == width )
	      col = 0, fmrow++, fmcol = v2._getArray();
       }
    return *this;
}

qmdMatrixRef&
qmdMatrixRef::setVectorScale( const qmdVectorRef& v, const qmdMatrixRef& m )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      getHeight() == v.getLength() &&
	      elementCount == m.elementCount &&
	      width == m.width );
#endif
    register reald *to = array;
    register reald *fmV = v._getArray();
    register reald *fmM = m.array;
    register reald val;
    int height = getHeight();
    for( int i = 0; i < height; i++ )
       {
	   val = *fmV++;
	   for( int j = 0; j < width; j++ )
	      (*to++) = val * (*fmM++);
       }
    return *this;
}

reald
qmdMatrixRef::getDeterminant3() const
{
#ifdef qmdDoBoundaryChecking
#endif
    qmdAssert( width == 3 && elementCount == 9 );
    reald d = 0.;
    d += array[0] * (array[4]*array[8] - array[5]*array[7] );
    d += array[1] * (array[5]*array[6] - array[3]*array[8] );
    d += array[2] * (array[3]*array[7] - array[4]*array[6] );
    return d;
}

reald
qmdMatrixRef::getDeterminant2() const
{
#ifdef qmdDoBoundaryChecking
#endif
    qmdAssert( width == 2 && elementCount == 4 );
    return( array[0]*array[3] - array[1]*array[2] );
}

reald
qmdMatrixRef::getDeterminant() const
{
#ifdef qmdDoBoundaryChecking
#endif
    switch( width )
       {
	  case 1:
	   return array[0];
	  case 2:
	   return getDeterminant2();
	  case 3:
	   return getDeterminant3();
	  default:
	    qmdAssert( elementCount == width*width );
	    qmdEigenStructure eigen( *this );
	    reald det = 1.;
	    for (int i=0; i<width; i++)
	      det *= eigen.getEigValues()(i);
	    return (det);
       }
}

ostream&
qmdMatrixRef::print( ostream& os, char *lineend ) const
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
#if 0
    os << "e= " << elementCount << ", w= " << width 
       << ", a= " << array << endl;
#endif
    for( ; count; count -= width )
       {
	   for( int i = 0; i < width; i++ )
	      os << " " << *rp++;
	   os << ((lineend)?lineend:"\0");
       }
    return os;
}

istream&
qmdMatrixRef::scan( istream& is )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       is >> *rp++;
    return is;
}

ostream&
qmdMatrixRef::write( ostream& os ) const
{
    return( os.write( (char *)array, elementCount * sizeof(reald) ) );
}

istream&
qmdMatrixRef::read( istream& is )
{
    return( is.read( (char *)array, elementCount * sizeof(reald) ) );
}

ostream&
operator<<( ostream& os, const qmdMatrixRef& qmd )
{
    return qmd.print( os );
}

istream&
operator>>( istream& is, qmdMatrixRef& qmd )
{
    return qmd.scan( is );
}

int
qmdMatrixRef::operator==( const qmdMatrixRef& op ) const
{
    if( elementCount != op.elementCount || width != op.width )
       return 0;
    for( int i = 0; i < elementCount; i++ )
       if( op.array[i] != array[i] )
	  return 0;
    return 1;
}

const qmdVectorRef&
qmdMatrixRef::getColumn( int i ) const
{
    static qmdVector column;
    int height = getHeight();
    column.set( height );
    for( int j = 0; j < height; j++ )
       column( j ) = array[ j*width + i ];
    return column;
}

const qmdVectorRef&
qmdMatrixRef::getRow( int i ) const
{
    static qmdVectorRef row;
    row.set( width, array+i*width );
    return row;
}

void
qmdMatrixRef::setColumn( int i, const qmdVectorRef& column )
{
    int height = getHeight();
    for( int j = 0; j < height; j++ )
       array[ j*width + i ] = column.getValue( j );
}

void
qmdMatrixRef::setRow( int i, const qmdVectorRef& row )
{
    qmdVectorRef( width, array+i*width ).copy( row );
}
