// qmdSubMatrixRef.c++

#include "qmdSubMatrixRef.h"
#include "qmdMatrixRef.h"
#include "qmdMatrix.h"
#include <memory.h>

qmdSubMatrixRef::qmdSubMatrixRef( const qmdMatrixRef& m, 
				int h, int w, int r, int c )
  : elementCount( ((h<=0)?m.getHeight():h) * ((w<=0)?m.getWidth():w) ),
    width( ((w<=0)?m.getWidth():w) ),
    spacing( m.getWidth() - w ),
    array( m._getArray(r,c) )
{}

qmdSubMatrixRef&
qmdSubMatrixRef::setData( reald *p )
{
    register reald *to = array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      *to++ = *p++;
	   to += spacing;
       }
    return *this;
}

reald *
qmdSubMatrixRef::getData( reald *p ) const
{
    register reald *fm = array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      *p++ = *fm++;
	   fm += spacing;
       }
    return p;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setAdd( const qmdSubMatrixRef& a, const qmdSubMatrixRef& b )
{
    reald *to = array;
    reald *fma = a.array;
    reald *fmb = b.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      (*to++) = (*fma++) + (*fmb++);
	   to += spacing;
	   fma += a.spacing;
	   fmb += b.spacing;
       }
    return *this;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setSubtract( const qmdSubMatrixRef& a, const qmdSubMatrixRef& b )
{
    reald *to = array;
    reald *fma = a.array;
    reald *fmb = b.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      (*to++) = (*fma++) - (*fmb++);
	   to += spacing;
	   fma += a.spacing;
	   fmb += b.spacing;
       }
    return *this;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setMultiply( const qmdSubMatrixRef& a, const qmdSubMatrixRef& b )
{
    register reald *to = array;
    register reald *fma;
    register reald *fmb;
    int height = elementCount / width;
    int bline = b.width + b.spacing;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      {
		  reald sum = 0.;
		  fma = a.getArray( i, 0 );
		  fmb = b.getArray( 0, j );
		  for( int k = 0; k < a.width; k++ )
		     sum += (*fma++) * (*fmb), fmb += bline;
		  *to++ = sum;
	      }
	   to += spacing;
       }
    return *this;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setMultiplyT( const qmdSubMatrixRef& a, const qmdSubMatrixRef& b )
{
    register reald *to = array;
    register reald *fma;
    register reald *fmb;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      {
		  reald sum = 0.;
		  fma = a.getArray( i, 0 );
		  fmb = b.getArray( j, 0 );
		  for( int k = 0; k < a.width; k++ )
		     sum += (*fma++) * (*fmb++);
		  *to++ = sum;
	      }
	   to += spacing;
       }
    return *this;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setAddEq( const qmdSubMatrixRef& a )
{
    reald *to = array;
    reald *fm = a.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       {
	   for( int j = 0; j < width; j++ )
	      (*to++) += (*fm++);
	   to += spacing;
	   fm += a.spacing;
       }
    return *this;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setInvert( const qmdSubMatrixRef& from )
{
    assert( getWidth() == from.getWidth() && getHeight() == from.getHeight() &&
	    getWidth() == getHeight() );
    
    qmdMatrix copy( elementCount/width, width );
    from.getData( copy.getArray() );
    copy.setInvertEq();
    setData( copy.getArray() );
    return *this;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setInvertEq()
{
    return setInvert( *this );
}

qmdSubMatrixRef&
qmdSubMatrixRef::setScaleEq( reald val )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       (*to++) *= val, to += spacing;
    return *this;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setConst( reald val )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       (*to++) = val, to += spacing;
    return *this;
}

qmdSubMatrixRef&
qmdSubMatrixRef::setDiagonal( reald val )
{
    int height = elementCount / width;
    int count = ( width < height ) ? width : height;
    setZero();
    while( count-- )
       (*this)( count, count ) = val;
    return *this;
}

ostream&
qmdSubMatrixRef::print( ostream& os, char *lineend )
{
    register int count = elementCount;
    register reald *fm = array;
    os << "e= " << elementCount << ", w= " << width 
       << ", a= " << array << endl;
    for( ; count; count -= width )
       {
	   for( int i = 0; i < width; i++ )
	      os << " " << *fm++, fm += spacing;
	   os << ((lineend)?lineend:"\0");
       }
    return os;
}

istream&
qmdSubMatrixRef::scan( istream& is )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       is >> *to++, to += spacing;
    return is;
}

ostream&
operator<<( ostream& os, qmdSubMatrixRef& sm )
{
    return sm.print( os );
}


qmdSubMatrixRef&
qmdSubMatrixRef::copy( const qmdSubMatrixRef& sm )
{
    assert( elementCount == sm.elementCount );
    width = sm.width;
    memcpy( array, sm.array, elementCount * sizeof(reald) );
    return *this;
}


ostream&
qmdSubMatrixRef::write( ostream& os )
{
    return( os.write( (char *)array, elementCount * sizeof(reald) ) );
}


istream&
qmdSubMatrixRef::read( istream& is )
{
    return( is.read( (char *)array, elementCount * sizeof(reald) ) );
}

