// qmdVectorRef.h++

#ifndef _qmdVectorRef_H_
#define _qmdVectorRef_H_

#include <stddef.h>
#include <iostream.h>
#include <assert.h>
#include <math.h>
#include "qmdreald.h"
#include "qmdMatrixRef.h"

class qmdVectorRef
{
    friend class qmdMatrixRef;
    friend class qmdVector;
    
protected:
    
    reald *array;
    int elementCount;
    
public:
    
    virtual ~qmdVectorRef() {}
    qmdVectorRef( )
      : array(0), elementCount(0) {}
    qmdVectorRef( int n, reald* a )
      : array(a), elementCount(n) {}
    qmdVectorRef( const qmdVectorRef& vr )
      : array( 0 ), elementCount( 0 ) { set( vr ); }

    qmdVectorRef& set( int n, reald* a );			// shadow
    qmdVectorRef& set( const qmdVectorRef& );		// shadow
    qmdVectorRef& copy( const qmdVectorRef& );		// copy
    qmdVectorRef& operator=( const qmdVectorRef& );	// shadow

    int getLength() const { return elementCount; }
    int getElementCount() const { return elementCount; }

    reald* getArray( int i = 0 ) { return( array + i ); }
    operator reald*() { return array; }
    reald& operator()( int i ) { return array[i]; }
    reald getValue( int i ) const { return array[i]; }
    inline qmdVectorRef subVector( int i, int n ) const;

    void setData( const reald *p );
    void getData( reald *p ) const;
    qmdVectorRef& operator<<( reald *p ) { setData(p); return *this; }
    void operator>>( reald *p ) { getData(p); }

    qmdVectorRef& setAdd( const qmdVectorRef&, const qmdVectorRef& );
    qmdVectorRef& setSubtract(const qmdVectorRef&, const qmdVectorRef& );
    qmdVectorRef& setNegate( const qmdVectorRef& );
    qmdVectorRef& setShift( reald, const qmdVectorRef& );
    qmdVectorRef& setScale( reald, const qmdVectorRef& );
    qmdVectorRef& setScaleAddEq( reald, const qmdVectorRef& );
    qmdVectorRef& setAddEq( const qmdVectorRef& op );
    qmdVectorRef& setSubtractEq( const qmdVectorRef& op );
    qmdVectorRef& setNegateEq();
    qmdVectorRef& setShiftEq( reald val );
    qmdVectorRef& setScaleEq( reald val );
    qmdVectorRef& setConst( reald val );
    qmdVectorRef& setZero() { return setConst( 0.); }
    qmdVectorRef& operator+=(const qmdVectorRef& op) {return setAddEq(op); }
    qmdVectorRef& operator-=(const qmdVectorRef& op) {return setSubtractEq(op);}
    qmdVectorRef& operator+=(const reald val) { return setShiftEq(val); }
    qmdVectorRef& operator-=(const reald val) { return setShiftEq(-val); }
    qmdVectorRef& operator*=(const reald val) { return setScaleEq(val); }
    qmdVectorRef& operator/=(const reald val) { return setScaleEq(reald(1)/val); }
    qmdVectorRef& operator=( const reald val ) { return setConst(val); }
    qmdVectorRef& setMultiply( const qmdMatrixRef&, const qmdVectorRef& );
    qmdVectorRef& setTMultiply( const qmdMatrixRef&, const qmdVectorRef& );
    qmdVectorRef& setVectorScale( const qmdVectorRef&, const qmdVectorRef& );
    reald getInnerProduct( const qmdVectorRef& ) const;
    int operator==( const qmdVectorRef& ) const;
    qmdVectorRef& setAbsolute();
    ostream& print( ostream& ) const;
    istream& scan( istream& );
    ostream& write( ostream& ) const;
    istream& read( istream& );
    qmdVectorRef& referTo( void* p ) { array = (reald*)p; return *this; }
    int getByteCount() { return( elementCount * sizeof(reald) ); }
    inline reald getMax( int* which = 0 ) const;
    inline reald getMin( int* which = 0 ) const;
    inline int setNormalize();
    qmdVectorRef& setCross( const qmdVectorRef&, const qmdVectorRef& );

    inline reald getSum() const;

private:
    inline reald* _getArray( int i=0 ) const { return( array + i ); }
    inline reald& _el( int i ) const { return array[i]; }
};

inline
qmdVectorRef
qmdVectorRef::subVector( int i, int n ) const
{
    return qmdVectorRef( n, array+i );
}

inline
reald
qmdVectorRef::getMax( int* which ) const
{
#ifdef qmdWarnings
    // should make sure array is valid !!
#endif
    reald maxv = array[0];
    int maxi = 0;
    for( int i = 1; i < elementCount; i++ )
       if( array[i] > maxv )
	  maxi = i, maxv = array[i];
    if( which )
       *which = maxi;
    return maxv;
}

inline
reald
qmdVectorRef::getMin( int* which ) const
{
#ifdef qmdWarnings
    // should make sure array is valid !!
#endif
    reald minv = array[0];
    int mini = 0;
    for( int i = 1; i < elementCount; i++ )
       if( array[i] < minv )
	  mini = i, minv = array[i];
    if( which )
       *which = mini;
    return minv;
}

inline
int
qmdVectorRef::setNormalize()
{
    reald mag = getInnerProduct( *this );
    mag = sqrt( mag );
    if( mag < .00001 )
       {
	   int which;
	   getMax( &which );
	   setZero();
	   array[which] = 1.;
	   return 1;
       }
    setScaleEq( 1./mag );
    return 0;
}

inline
reald
qmdVectorRef::getSum() const
{
    reald sum = 0;
    for( int i = 0; i < elementCount; i++ )
       sum += array[i];
    return sum;
}


extern ostream& operator<<( ostream&, const qmdVectorRef& );
extern istream& operator>>( istream&, qmdVectorRef& );

#endif /* _qmdVectorRef_H_ */
