// qmGaussianInvRef.c++

#include <math.h>
#include "qmGaussianInvRef.h"

real qmGaussianInvRef::MinDet = 1.e-30;

ostream&
qmGaussianInvRef::alert( ostream& os, char* msg )
{
    return( os << "qmGaussianInvRef: ALERT: " << msg << endl );
}

void
qmGaussianInvRef::set( const int dim, real* a )
{
    mean.set( dim, a );
    invCovariance.set( dim, dim, a+dim );
    state = NOT_INITED; // ??
}

void
qmGaussianInvRef::set( const qmGaussianInvRef& g )
{
    state = g.state;
    mean = g.mean;
    invCovariance = g.invCovariance;
}

void
qmGaussianInvRef::copy( const qmGaussianInvRef& g )
{
    state = g.state;
    mean.copy( g.mean );
    invCovariance.copy( g.invCovariance );
}

void
qmGaussianInvRef::setInvCovariance( const qmMatrixRef& m )
{
    invCovariance.copy( m ); 
    testSetSingular();
}

inline
real
qmGaussianInvRef::_getMahalanobisSquare( const qmVectorRef& v ) const
{
    int n = v.getLength();
    qmVector dif( n );
    qmVector temp( n );
    dif.setSubtract( v, mean );
    return( dif.getInnerProduct( temp.setMultiply( invCovariance, dif ) ) );
}

real 
qmGaussianInvRef::getMahalanobisSquare( const qmVectorRef& v ) const
{
#ifdef QmStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular invCovariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobisSquare: object is not initialized" );
    if( SINGULAR == state )
       if( v == mean )
	  return 0.;
       else
	  return QmMaxReal;
    return _getMahalanobisSquare( v );
}

real 
qmGaussianInvRef::getMahalanobis( const qmVectorRef& v ) const
{
#ifdef QmStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular invCovariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobis: object is not initialized" );
    if( SINGULAR == state )
       return 0.;
    return sqrt( _getMahalanobisSquare( v ) );
}

inline
real
qmGaussianInvRef::_getLogLikelihood( const qmVectorRef& v ) const
{
    // i am not sure this is right
    return( (-0.5)*( getDim() * (real)log(2*M_PI) -
		     (real)log( invCovariance.getDeterminant() ) + 
		     _getMahalanobisSquare( v ) ) );
}

real 
qmGaussianInvRef::getLogLikelihood( const qmVectorRef& v ) const
{
#ifdef QmStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular invCovariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobis: object is not initialized" );
    if( SINGULAR == state )
       return -QmMaxReal;
    return _getLogLikelihood( v );
}

real 
qmGaussianInvRef::getLikelihood( const qmVectorRef& v ) const
{
#ifdef QmStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular invCovariance!" );
#endif
    if( SINGULAR == state )
       return 0.;
    return( (real) exp( _getLogLikelihood( v ) ) );
}

ostream&
qmGaussianInvRef::write( ostream& os )
{
    mean.write( os );
    invCovariance.write( os );
    return os;
}

istream& 
qmGaussianInvRef::read( istream& is )
{
    mean.read( is );
    invCovariance.read( is );
    return is;
}

ostream& 
qmGaussianInvRef::printPretty( ostream& os, char* lineend )
{
    os << "state= " << state << endl;
    mean.print( os << "m= " ) << endl;
    invCovariance.print( os << "invcov= ", lineend ) << endl;
    return os;
}

//istream& operator>>( istream& is, qmGaussianInvRef& g )
// { return g.scan( is ); }
//ostream& operator<<( ostream& os, qmGaussianInvRef& g )
// { return g.print( os );}
