// qmGaussianRef.c++

#include <math.h>
#include "qmGaussianRef.h"

real qmGaussianRef::MinDet = 1.e-30;

ostream&
qmGaussianRef::alert( ostream& os, char* msg )
{
    return( os << "qmGaussianRef: ALERT: " << msg << endl );
}

void
qmGaussianRef::set( const int dim, real* a )
{
    mean.set( dim, a );
    covariance.set( dim, dim, a+dim );
    state = NOT_INITED;
}

void
qmGaussianRef::set( const qmGaussianRef& g )
{
    state = g.state;
    mean = g.mean;
    covariance = g.covariance;
}

void
qmGaussianRef::copy( const qmGaussianRef& g )
{
    state = g.state;
    mean.copy( g.mean );
    covariance.copy( g.covariance );
}

void
qmGaussianRef::setCovariance( const qmMatrixRef& m )
{
    covariance.copy( m ); 
    if( covariance.getDeterminant() < MinDet )
       state = SINGULAR;
    else
       state = NONSINGULAR;
}

inline
real
qmGaussianRef::_getMahalanobisSquare( const qmVectorRef& v ) const
{
    int n = v.getLength();
    qmVector dif( n );
    qmVector temp( n );
    qmMatrix inv_cov( covariance );
    inv_cov.setInvertEq();
    dif.setSubtract( v, mean );
    return( dif.getInnerProduct( temp.setMultiply( inv_cov, dif ) ) );
}

real 
qmGaussianRef::getMahalanobisSquare( const qmVectorRef& v ) const
{
#ifdef QmStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular covariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobisSquare: object is not initialized" );
    if( SINGULAR == state )
       if( v == mean )
	  return 0.;
       else
	  return QmMaxReal;
    return _getMahalanobisSquare( v );
}

real 
qmGaussianRef::getMahalanobis( const qmVectorRef& v ) const
{
#ifdef QmStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular covariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobis: object is not initialized" );
    if( SINGULAR == state )
       if( v == mean )
	  return 0.;
       else
	  return QmMaxReal;
    return sqrt( _getMahalanobisSquare( v ) );
}

inline
real
qmGaussianRef::_getLogLikelihood( const qmVectorRef& v ) const
{
    return( (-0.5)*( getDim() * (real)log(2*M_PI) +
		     (real)log( covariance.getDeterminant() ) + 
		     _getMahalanobisSquare( v ) ) );
}

real 
qmGaussianRef::getLogLikelihood( const qmVectorRef& v ) const
{
#ifdef QmStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular covariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobis: object is not initialized" );
    if( SINGULAR == state )
       if( v == mean )
	  return QmMaxReal;
       else
	  return -QmMaxReal;
    return _getLogLikelihood( v );
}

real 
qmGaussianRef::getLikelihood( const qmVectorRef& v ) const
{
#ifdef QmStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular covariance!" );
#endif
    if( SINGULAR == state )
       if( v == mean )
	  return QmMaxReal;
       else
	  return 0.;
    return( (real) exp( _getLogLikelihood( v ) ) );
}

ostream&
qmGaussianRef::write( ostream& os )
{
    mean.write( os );
    covariance.write( os );
    return os;
}

istream& 
qmGaussianRef::read( istream& is )
{
    mean.read( is );
    covariance.read( is );
    return is;
}

ostream& 
qmGaussianRef::printPretty( ostream& os, char* lineend )
{
    os << "state= " << state << endl;
    mean.print( os << "m= " ) << endl;
    covariance.print( os << "cov= ", lineend ) << endl;
    return os;
}

//istream& operator>>( istream& is, qmGaussianRef& g ) { return g.scan( is ); }
//ostream& operator<<( ostream& os, qmGaussianRef& g ) { return g.print( os );}
