// qmdGaussianInvRef.c++

#include <math.h>
#include "qmdGaussianInvRef.h"

reald qmdGaussianInvRef::MinDet = 1.e-30;

ostream&
qmdGaussianInvRef::alert( ostream& os, char* msg )
{
    return( os << "qmdGaussianInvRef: ALERT: " << msg << endl );
}

void
qmdGaussianInvRef::set( const int dim, reald* a )
{
    mean.set( dim, a );
    invCovariance.set( dim, dim, a+dim );
    state = NOT_INITED; // ??
}

void
qmdGaussianInvRef::set( const qmdGaussianInvRef& g )
{
    state = g.state;
    mean = g.mean;
    invCovariance = g.invCovariance;
}

void
qmdGaussianInvRef::copy( const qmdGaussianInvRef& g )
{
    state = g.state;
    mean.copy( g.mean );
    invCovariance.copy( g.invCovariance );
}

void
qmdGaussianInvRef::setInvCovariance( const qmdMatrixRef& m )
{
    invCovariance.copy( m ); 
    testSetSingular();
}

inline
reald
qmdGaussianInvRef::_getMahalanobisSquare( const qmdVectorRef& v ) const
{
    int n = v.getLength();
    qmdVector dif( n );
    qmdVector temp( n );
    dif.setSubtract( v, mean );
    return( dif.getInnerProduct( temp.setMultiply( invCovariance, dif ) ) );
}

reald 
qmdGaussianInvRef::getMahalanobisSquare( const qmdVectorRef& v ) const
{
#ifdef qmdStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular invCovariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobisSquare: object is not initialized" );
    if( SINGULAR == state )
       if( v == mean )
	  return 0.;
       else
	  return qmdMaxReal;
    return _getMahalanobisSquare( v );
}

reald 
qmdGaussianInvRef::getMahalanobis( const qmdVectorRef& v ) const
{
#ifdef qmdStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular invCovariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobis: object is not initialized" );
    if( SINGULAR == state )
       return 0.;
    return sqrt( _getMahalanobisSquare( v ) );
}

inline
reald
qmdGaussianInvRef::_getLogLikelihood( const qmdVectorRef& v ) const
{
    // i am not sure this is right
    return( (-0.5)*( getDim() * (reald)log(2*M_PI) -
		     (reald)log( invCovariance.getDeterminant() ) + 
		     _getMahalanobisSquare( v ) ) );
}

reald 
qmdGaussianInvRef::getLogLikelihood( const qmdVectorRef& v ) const
{
#ifdef qmdStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular invCovariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobis: object is not initialized" );
    if( SINGULAR == state )
       return -qmdMaxReal;
    return _getLogLikelihood( v );
}

reald 
qmdGaussianInvRef::getLikelihood( const qmdVectorRef& v ) const
{
#ifdef qmdStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular invCovariance!" );
#endif
    if( SINGULAR == state )
       return 0.;
    return( (reald) exp( _getLogLikelihood( v ) ) );
}

ostream&
qmdGaussianInvRef::write( ostream& os )
{
    mean.write( os );
    invCovariance.write( os );
    return os;
}

istream& 
qmdGaussianInvRef::read( istream& is )
{
    mean.read( is );
    invCovariance.read( is );
    return is;
}

ostream& 
qmdGaussianInvRef::printPretty( ostream& os, char* lineend )
{
    os << "state= " << state << endl;
    mean.print( os << "m= " ) << endl;
    invCovariance.print( os << "invcov= ", lineend ) << endl;
    return os;
}

//istream& operator>>( istream& is, qmdGaussianInvRef& g )
// { return g.scan( is ); }
//ostream& operator<<( ostream& os, qmdGaussianInvRef& g )
// { return g.print( os );}
