// qmdGaussianInvRef.h++

#ifndef _qmdGaussianInvRef_H_
#define _qmdGaussianInvRef_H_

#include "QuickMathDouble.h"

class qmdGaussianInvRef
{
    static ostream& alert( ostream&, char* msg );
    inline reald _getMahalanobisSquare( const qmdVectorRef& ) const;
    inline reald _getLogLikelihood( const qmdVectorRef& ) const;
protected:
    inline int _testSingular() const;
    inline void testSetSingular();

    unsigned int state;
    qmdVectorRef mean;
    qmdMatrixRef invCovariance;

public:

    static reald MinDet;  // distribution considered singular if det < MinDet
    static int getStorageSize( int n ) { return( sizeof(reald)*n*(n+1) ); }
    enum { NONSINGULAR=0, SINGULAR=1, NOT_INITED=2, NAN=3 }; // NAN not impl

    qmdGaussianInvRef()
      : state( NOT_INITED ), mean( ), invCovariance( ) {}
    qmdGaussianInvRef( const int dim, reald* a )
      : state( NOT_INITED ), mean( dim, a ), invCovariance( dim, dim, a+dim ){}
    qmdGaussianInvRef( const qmdGaussianInvRef& g )
      : state( g.state ), mean( g.mean ), invCovariance( g.invCovariance ) {}
    qmdGaussianInvRef( const qmdVectorRef& v, const qmdMatrixRef& m )
      : state( NOT_INITED ), mean( v ), invCovariance( m )
    { testSetSingular();}

    void set( const int dim, reald* a );
    void set( const qmdGaussianInvRef& );
    void copy( const qmdGaussianInvRef& );
			  
    inline int getDim() const { return mean.getLength(); }
    const qmdVectorRef& getMean() const { return mean; } 
    void setMean( const qmdVectorRef & v ) { mean.copy( v ); }
    const qmdMatrixRef& getInvCovariance() const { return invCovariance; }
    void setInvCovariance( const qmdMatrixRef& m );
    inline void set( const qmdVectorRef&, const qmdMatrixRef& );
    int getState() const { return state; }
    int isSingular() const { return( SINGULAR == state ); }

    reald getMahalanobisSquare( const qmdVectorRef& ) const;
    reald getMahalanobis( const qmdVectorRef& ) const;
    reald getLikelihood( const qmdVectorRef& ) const;
    reald getLogLikelihood( const qmdVectorRef& ) const;

    inline int  cull( const qmdVectorRef&, reald alpha ); 

    ostream& write( ostream& );
    istream& read( istream& );
    ostream& printPretty( ostream&, char* = "\n" );
    // istream& scan( istream& );

    inline void setSingular();
};

// do a liner distance and crop at alpha
// 1 = cull it
// 0 = within cull boundry - don't cull
inline
int
qmdGaussianInvRef::cull( const qmdVectorRef& v, reald threshold )
{
    int cullp = 0;
    int n = v.getLength();
    qmdVector dif( n );
    dif.setSubtract( v, mean );
    
    for (int i=0; i<n; i++)
      if ( threshold < (invCovariance(i,i) * fabs(dif(i))) )
	cullp = 1;
    
    return( cullp );
}

inline
void
qmdGaussianInvRef::set( const qmdVectorRef& v, const qmdMatrixRef& m )
{
    setMean( v );
    setInvCovariance( m );
}

inline
int
qmdGaussianInvRef::_testSingular() const
{
    if( invCovariance.getDeterminant() < MinDet )
       return 1;
    return 0;
}

inline
void
qmdGaussianInvRef::testSetSingular()
{
    if( _testSingular() )
       state = SINGULAR;
    else
       state = NONSINGULAR;
}

inline
void
qmdGaussianInvRef::setSingular()
{
  state = SINGULAR;
}

#endif /* _qmdGaussianInvRef_H_ */
