// qmdGaussianRef.c++

#include <math.h>
#include "qmdGaussianRef.h"

reald qmdGaussianRef::MinDet = 1.e-30;

ostream&
qmdGaussianRef::alert( ostream& os, char* msg )
{
    return( os << "qmdGaussianRef: ALERT: " << msg << endl );
}

void
qmdGaussianRef::set( const int dim, reald* a )
{
    mean.set( dim, a );
    covariance.set( dim, dim, a+dim );
    state = NOT_INITED;
}

void
qmdGaussianRef::set( const qmdGaussianRef& g )
{
    state = g.state;
    mean = g.mean;
    covariance = g.covariance;
}

void
qmdGaussianRef::copy( const qmdGaussianRef& g )
{
    state = g.state;
    mean.copy( g.mean );
    covariance.copy( g.covariance );
}

void
qmdGaussianRef::setCovariance( const qmdMatrixRef& m )
{
    covariance.copy( m ); 
    if( covariance.getDeterminant() < MinDet )
       state = SINGULAR;
    else
       state = NONSINGULAR;
}

inline
reald
qmdGaussianRef::_getMahalanobisSquare( const qmdVectorRef& v ) const
{
    int n = v.getLength();
    qmdVector dif( n );
    qmdVector temp( n );
    qmdMatrix inv_cov( covariance );
    inv_cov.setInvertEq();
    dif.setSubtract( v, mean );
    return( dif.getInnerProduct( temp.setMultiply( inv_cov, dif ) ) );
}

reald 
qmdGaussianRef::getMahalanobisSquare( const qmdVectorRef& v ) const
{
#ifdef qmdStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular covariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobisSquare: object is not initialized" );
    if( SINGULAR == state )
       if( v == mean )
	  return 0.;
       else
	  return qmdMaxReal;
    return _getMahalanobisSquare( v );
}

reald 
qmdGaussianRef::getMahalanobis( const qmdVectorRef& v ) const
{
#ifdef qmdStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular covariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobis: object is not initialized" );
    if( SINGULAR == state )
       if( v == mean )
	  return 0.;
       else
	  return qmdMaxReal;
    return sqrt( _getMahalanobisSquare( v ) );
}

inline
reald
qmdGaussianRef::_getLogLikelihood( const qmdVectorRef& v ) const
{
    return( (-0.5)*( getDim() * (reald)log(2*M_PI) +
		     (reald)log( covariance.getDeterminant() ) + 
		     _getMahalanobisSquare( v ) ) );
}

reald 
qmdGaussianRef::getLogLikelihood( const qmdVectorRef& v ) const
{
#ifdef qmdStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular covariance!" );
#endif
    if( NOT_INITED == state )
       alert( cerr, "getMahalanobis: object is not initialized" );
    if( SINGULAR == state )
       if( v == mean )
	  return qmdMaxReal;
       else
	  return -qmdMaxReal;
    return _getLogLikelihood( v );
}

reald 
qmdGaussianRef::getLikelihood( const qmdVectorRef& v ) const
{
#ifdef qmdStatisticsWarnings
    if( SINGULAR == state )
       alert( cerr, "getMahalanobisSquare: singular covariance!" );
#endif
    if( SINGULAR == state )
       if( v == mean )
	  return qmdMaxReal;
       else
	  return 0.;
    return( (reald) exp( _getLogLikelihood( v ) ) );
}

ostream&
qmdGaussianRef::write( ostream& os )
{
    mean.write( os );
    covariance.write( os );
    return os;
}

istream& 
qmdGaussianRef::read( istream& is )
{
    mean.read( is );
    covariance.read( is );
    return is;
}

ostream& 
qmdGaussianRef::printPretty( ostream& os, char* lineend )
{
    os << "state= " << state << endl;
    mean.print( os << "m= " ) << endl;
    covariance.print( os << "cov= ", lineend ) << endl;
    return os;
}

//istream& operator>>( istream& is, qmdGaussianRef& g ) { return g.scan( is ); }
//ostream& operator<<( ostream& os, qmdGaussianRef& g ) { return g.print( os );}
