// qmdKMeans.c++

#include "qmdKMeans.h"

qmdKMeans::~qmdKMeans()
{
    if( clusters ) delete [] clusters;
    if( clusterData ) delete [] clusterData;
    if( sampleMap ) delete [] sampleMap;
    if( samples ) delete [] samples;
    if( sampleWeights ) delete [] sampleWeights;
    if( sampleData ) delete [] sampleData;
}

void
qmdKMeans::setDimension( int dim )
{
    if( dim > 0 )
       dimension = dim;
    if( clusterCount )
       // have to resize clusterData and elements of clusters
       {
	   if( clusterData )
	      delete [] clusterData;
	   clusterData = new reald[ dim * clusterCount ];
	   for( int i = 0; i < clusterCount; i++ )
	      clusters[i].set( dim, clusterData + i*dim );
       }
    if( sampleCount )
       // have to resize sampleData and elements of samples;
       {
	   if( sampleData )
	      delete [] sampleData;
	   sampleData = new reald[ dim * sampleCount ];
	   for( int i = 0; i < sampleCount; i++ )
	      samples[i].set( dim, sampleData + i*dim );
       }
}

void
qmdKMeans::setClusterCount( int M )
{
    if( M > 0 )
       clusterCount = M;
    if( clusters )
       delete [] clusters;
    if( clusterValid )
       delete [] clusterValid;
    clusters = new qmdVectorRef[ M ];
    clusterValid = new int[ M ];
    if( dimension )
       {
	   if( clusterData )
	      delete [] clusterData;
	   clusterData = new reald[ dimension * clusterCount ];
	   for( int i = 0; i < clusterCount; i++ )
	      clusters[i].set( dimension, clusterData + i*dimension );
       }
}

void
qmdKMeans::setSampleCount( int N )
{
    if( N <= 0 )
       return;
    if( sampleCount != N )
       {
	   if( sampleCount )
	      {
		  delete [] sampleMap;
		  delete [] sampleWeights;
      delete [] samples;
	      }
	   sampleMap = new int[ N ];
	   sampleWeights = new reald[ N ];
	   samples = new qmdVectorRef[ N ];
	   sampleCount = N;
       }
    if( dimension )
       {
	   if( sampleData )
	      delete [] sampleData;
	   sampleData = new reald[ dimension * sampleCount ];
	   for( int i = 0; i < sampleCount; i++ )
	      samples[i].set( dimension, sampleData + i*dimension );
       }
}

void
qmdKMeans::setClusterMean( int i, const qmdVectorRef& m )
{
    if( !inited() || i < 0 || i >= clusterCount )
       return;
    clusters[i].copy( m );
}

void
qmdKMeans::setSample( int i, const qmdVectorRef& s )
{
    if( !inited() || i < 0 || i >= sampleCount )
       return;
    samples[i].copy( s );
    sampleMap[i] = -1;
}

void
qmdKMeans::setSampleWeight( int i, reald w )
{
    if( !inited() || i < 0 || i >= sampleCount )
       return;
    sampleWeights[i] = w;
}

void
qmdKMeans::cluster()
{
    for( int i = 0; i < sampleCount; i++ )
       sampleMap[i] = -1;
    // for( i = 0; i < sampleCount; i++ )
    // samples[i].print( cerr << "\t" ) << endl;
    while( reclassify() )
       for( int j = 0; j < clusterCount; j++ )
	  updateClusterMean( j );
}

static 
inline
reald
dist( qmdVectorRef& a, qmdVectorRef& b )
{
    qmdVector diff( a );
    diff.setSubtractEq( b );
    return diff.getInnerProduct( diff );
}

int
qmdKMeans::findClosestClass( qmdVectorRef& s )
{
    reald min = dist( s, clusters[0] );
    int imin = 0;
    reald d;
    for( int i = 1; i < clusterCount; i++ )
       {
	   d = dist( s, clusters[i] );
	   if( d < min )
	      min = d, imin = i;
       }
    return imin;
}

ostream&
qmdKMeans::printClassification( ostream& os )
{
    for( int i = 0; i < sampleCount; i++ )
       os << sampleMap[i] << "  ";
    return os;
}

int
qmdKMeans::reclassify()
{
    // printClassification( cerr << "[  " ) << "] ";
    int changes = 0;
    for( int i = 0; i < sampleCount; i++ )
       {
	   int newClass = findClosestClass( samples[i] );
	   if( newClass != sampleMap[i] )
	      changes++;
	   sampleMap[i] = newClass;
       }
    // printClassification( cerr << "--> [  " ) << "]" << endl;
    return changes;
}

void
qmdKMeans::updateClusterMean( int i )
{
    clusters[i].setZero();
    reald totalWeight = 0.;
    clusterValid[i] = 0;
    for( int j = 0; j < sampleCount; j++ )
       if( sampleMap[j] == i )
	  {
	      clusters[i].setScaleAddEq( sampleWeights[j], samples[j] );
	      totalWeight += sampleWeights[j];
	      clusterValid[i]++;
	  }
    clusters[i] /= totalWeight;
}

ostream&
qmdKMeans::print( ostream& os )
{
    os << "qmdKMeans:" << endl;
    os << "\tdim= " << dimension << endl;
    os << "\tNsamp= " << sampleCount << endl;
    for( int i = 0; i < sampleCount; i++ )
       os << "\t\t[" << sampleMap[i] << "][" << sampleWeights[i]
	  << "] " << samples[i] << endl;
    os << "\tNclus= " << clusterCount << endl;
    for( i = 0; i < clusterCount; i++ )
       os << "\t\t[" << i << "] " << clusters[i] << endl;
    return os;
}
